//=============================================================================
// XCoop.
//=============================================================================
class XCoop expands CoopGame
	config(xCoop);

Struct C
{
	var() config string OriginalClass;
	var() config string ReplacementClass;
};

struct MapTable
{
	Var() config string MapName;
	var() config string NextMapOverride;
	var() config Byte 	MapTimerMinutes;
};

Struct Scoretable
{
	var() config string	 EnemyName;
	var() config int	 PointYield;
};

Struct As
{
	var() config name 	Command;
	var() config int 	MinAdminLvl;
};

Enum EPunishmentType
{
	PT_Kill, PT_Kick
};

Struct HUDColor
{
	Var() config Color SayColor,ServerEventColor,ChatlinkColor,PSayColor,ASayColor;	
};

Struct S
{
	Var() config String LevelEndSnd,Spree,Dominating,DoubleKill,Godlike,KillingSpree,MonsterKill,Multikill,Rampage,UltraKill,Unstoppable,WickedSick;
	var() config bool bUseSprees;
};

Struct FF
{
	var() config bool bNullifyFFDamage;
	var() config bool bReportFF;
	var() config Float FFDamageScaling;
};

Struct RS
{
	Var() config bool bRespawnItems;
	var() config Int RespawnTime;
	var() config Bool bCoopWeaponMode;
};

Struct PGI
{
	var() Config String PlayerPassword;
};

Struct PGiveItems
{
	var() config string PGiveitem[16];
};

//=============================================
// Editable variablez
//=============================================
var(HookMutator) config String HookMutator[32];

var(gamecontrol) config bool bShowEnds;
var(GameControl) config bool bPlayersCanAbortEnding;
var(Gamecontrol) config Scoretable Scoreyieldtable[128];
Var(GameControl) Config s SpreeSettings;
var(GameControl) config FF FriendlyFireSettings;
var(GameControl) config RS ItemRespawning;
var(GameControl) Config String GiveItem[64];
var(GameControl) config Bool bKeepItemsAfterDeath;
var(GameControl) config bool bInstantRespawnAfterDeath;
var(GameControl) config byte GameDifficulty;
var(GameControl) config string PlayerReplacementName;
var(GameControl) config bool bAntiSpam;
var(GameControl) Config string StartupCommand[32];
var(GameControl) config string LeftMessage,EnteredMessage;
var(GameControl) config float MinEndAbortPercentage;

var(HUDControl) config Byte TalktexScale;
var(HUDControl) config string Logotex;
var(HUDControl) config string DefaultTalkTexture;
var(HUDControl) config Class<texture> DefaultTalkTexClass;
var(HUDControl) config Class<texture> LogoClass;
var(HUDControl) config HUDColor MessageColors;
var(HUDControl) config bool bQuickDraw;
var(HUDControl) config ERenderstyle Logostyle;

var(SequenceControl)	config MapTable SeqTable[128];
var(SequenceControl) 	config bool bOverrideMapEnds;
var(SequenceControl) 	config bool bUseMaptimers;

var(playercontrol) config bool bNoCustomPlayerpawns;		// Players can't login with class=..
var(PlayerControl) config bool bOnlyXPlayerpawns;			// Only xPlayers allowed.
var(PlayerControl) config C ClassReplacement[32];
var(PlayerControl) config string RestrictedPlayerclass[32];

var(Handlers) config String CommandMod[4];
var(Handlers) config String LoginHandlerClass;
var LoginHandler LoginHandler;
var(Handlers) config string AuthorityHandlerClass;
var AuthorityHandler AuthorityHandler;
var(Handlers) config string MessageHandlerClass;
var MessageHandler MessageHandler;
var(Handlers) config string SummonHandlerClass;
var SummonHandler Summonhandler;

var(LoginControl) private config string AdminPassword[3];
var(LoginControl) private config string GamePassword;

var(AdminControl) config As AuthorityTable[64];			// Table that outlines who can perform what commands.
var(admincontrol) config bool bAnnounceAdminCommands;
var(AdminControl) config String RestrictedKeyword[64];

//==============
// Other varz
//==============
var XGRI XGRI;
var MapTimer MapTimer;
var string LevelName;
var int EndCount,NoCount;
var PlayerPawn EndingPlayer;
var pawn Temp;
var string EndURL;
var bool bAdvanced;
var botinfo botconfig;
var bool bCanEnd;
var actor a;
var texture DefaultTalkTex;

var class<xCommands> CMDList[4];
var class<inventory> GiveList[64];
var int InvCount;

var scorekeeper scorekeeper;

Var Sound LevelEndSnd,Spree,Dominating,DoubleKill,Godlike,KillingSpree,MonsterKill,Multikill,Rampage,UltraKill,Unstoppable,WickedSick;

//===================================
// Handler spawning / Startup
//===================================
Auto State active
{
	Begin:
	sleep(0.0);
	bCanEnd=true;
	LevelName = GETURLMAP();
	SwapTeleporters();
	SetPRIVars();
	If( EnteredMessage == "" )
		EnteredMessage = " entered the game.";
	If( LeftMessage == "" )
		LeftMessage = " left the game.";
	
	Idle:
}

event InitGame( string Options, out string Error )
{
	local string InOpt, LeftOpt;
	local int pos;
	local class<Mutator> MClass;
	local int i;

	log( "InitGame:" @ Options );
	MaxPlayers = GetIntOption( Options, "MaxPlayers", MaxPlayers );
	InOpt = ParseOption( Options, "Difficulty" );
	Difficulty = GameDifficulty;
	log( "Difficulty" @ Difficulty );
	
	LoadHookmutators();
	GetItems();
	LoadCommands();

	for(i=0;i<3;i++)
	log( "Remote Administration with Password" @ AdminPassWord[i] );		// DO NOT LET INIT GAME CHANGE PASSWORDS!!

	InOpt = ParseOption( Options, "GameSpeed");
	if( InOpt != "" )
	{
		log("GameSpeed"@InOpt);
		SetGameSpeed(float(InOpt));
	}

	InOpt = ParseOption( Options, "GamePassword");
	if( InOpt != "" )
	{
		GamePassWord = InOpt;
		log( "GamePassword" @ InOpt );
	}
	
	For( I=0; I < 32; i++ )
	{
		If( StartupCommand[i] != "" )
		{
			Log("AutoCommand: "$Startupcommand[i],'xCoop');
			Consolecommand(startupcommand[i]);
		}
	}

	InOpt = ParseOption( Options, "LocalLog");
	if( InOpt ~= "true" )
		bLocalLog = True;

	InOpt = ParseOption( Options, "WorldLog");
	if( InOpt ~= "true" )
		bWorldLog = True;
	CurrentID = 1;
	Scorekeeper = New(Self,'Scorekeeper') class'scorekeeper';
	ScoreKeeper.Game=Self;
}

function PreBeginPlay()
{
	StartTime = 0;
	SetTimer(1.0, true);
	SetGameSpeed(GameSpeed);
	Level.bNoCheating = bNoCheating;
	bCoopWeaponMode=ItemRespawning.bCoopweaponmode;
	
	if (GameReplicationInfoClass != None)
	{
		GameReplicationInfo = Spawn(GameReplicationInfoClass);
		XGRI = XGRI(Gamereplicationinfo);
		XGRI.Saycolor=MessageColors.Saycolor;
		XGRI.ChatlinkColor=Messagecolors.ChatlinkColor;
		XGRI.ServerEventColor=Messagecolors.ServerEventColor;	// booya!
		XGRI.PSAYCOLOR = Messagecolors.PSAYCOLOR;
		XGRI.ASAYCOLOR = Messagecolors.ASAYCOLOR;
		XGRI.Talktexscale = TalkTexScale;

		if(Logotex!="")
		{
			XGRI.LogoTex=Texture( dynamicloadobject( LogoTex, LogoClass ) );
			XGRI.bQuickDraw = bQuickDraw;
			XGRI.LogoStyle = LogoStyle;
		}
		if(DefaultTalkTexture!="" && DefaultTalkTexClass!=None)
		{
			DefaultTalkTex = Texture( DynamicLoadObject( DefaultTalkTexture, Class'Texture' ) );
		}
	}
	else
		GameReplicationInfo = Spawn(class'GameReplicationInfo');
	InitGameReplicationInfo();
}	

Function GetItems()
{
local int i;
local Class<inventory> inv;
InvCount=0;
	For(i=0;i<64;i++)
	{
		if( Giveitem[i]!="" )
		{
			Inv = Class<Inventory>(DynamicLoadObject(GiveItem[i],class'class'));
			if( Inv==none )
			{
				Log("GiveItem: Failed to load inventory class "$Giveitem[i],'xCoop');
				Continue;
			}
			GiveList[InvCount]=Inv;
			InvCount++;
		}
	}
}

function LoadCommands()
{
local int i,b;
local Class<xCommands> inv;
b=0;
	For(i=0;i<4;i++)
	{
		if( CommandMod[i]!="" )
		{
			Inv = Class<xCommands>(DynamicLoadObject(CommandMod[i],class'class'));
			if( Inv==none )
			{
				Log("CommandList: Failed to load inventory class "$CommandMod[i],'xCoop');
				Continue;
			}
			CMDList[b]=Inv;
			b++;
		}
	}
}

Function LoadHookMutators()
{
local int i,H;
local class<mutator> Mut;
local Mutator M;
	
	For(i=0;i<32;i++)
	{
		If( Hookmutator[i] == "" ) continue;
		Mut = Class<Mutator>(Dynamicloadobject(HookMutator[i],class'Class'));	
		
		If( Mut==none )
		{
			Log("HookMutator: Failed to load "$Mut$".",'xCoop');
			Continue;
		}
		M=Spawn(Mut);
		If( M==None )
		{
			Log("HookMutator: Failed to create "$Mut$".",'xCoop');
			Continue;
		}
		
		If( BaseMutator == None ) 
		{
			BaseMutator = M;	
		}
		Else
		{
			BaseMutator.AddMutator( M );
		}
	}
	If( BaseMutator == None )
		BaseMutator = Spawn(class'Mutator');
}
	
Function SwapTeleporters()
{
local int i;
local teleporter t;
local navigationpoint N;
local bool bOver;

	If( bOverrideMapends || bUseMapTimers )
	{
		For(I=0;I<128;i++)
		{
			If( SeqTable[i].MapName ~= LevelName )
			{
				If( bUseMapTimers )
				{
					MapTimer = Spawn(class'MapTimer');
					MapTimer.XCoop=Self;
					If( SeqTable[i].Maptimerminutes > 0 )
					{
						MapTimer.Minutes = SeqTable[i].Maptimerminutes;
						Log(LevelName@"set to end in"@Maptimer.Minutes@"minutes.",'xCoop');
					}
					Else
					{
						MapTimer.Destroy();
						Log("This map doesn't have a map timer set.",'xCoop');
					}
				}
				If( bOverrideMapEnds )
				bOver=True;
				Break;
			}
		}
	}

	For( N = Level.NavigationPointList; N != None; N = N.NextNavigationPoint )
	{
		If( N.IsA('Teleporter') )
			T = Teleporter(N);
		Else Continue;
				
		If( InStr( T.URL, "/" ) >= 0 || InStr( T.URL, "#" ) >= 0 )
		{
			If( bshowends )
			{
				T.bHidden=False;
				t.bAlwaysRelevant=True;
			}
			If( bOver )
			{
				Log("Next map override:"@T@"had its URL modified and it now points to the map"@SeqTable[i].NextMapOverride,'xCoop');
				T.URL = seqtable[i].nextmapoverride$"#NextMapOverride?peer";
			}
		}
	}
}

Function SetPRIVars()
{
	Consolecommand("Set Bots PlayerReplicationInfoClass XCoop.XPRI");	
}

Function PostBeginPlay()
{
	Super.postbeginplay();
	BotConfig = spawn(class'BotInfo');
	LoadHandlers();
	If( SpreeSettings.bUseSprees )
	{
		// Load sounds.
		LevelEndSnd = Sound( DynamicLoadobject( SpreeSettings.LevelEndSnd, Class'Sound' ) );
		Spree = Sound( DynamicLoadobject( SpreeSettings.Spree, Class'Sound' ) );
		Dominating = Sound( DynamicLoadobject( SpreeSettings.Dominating, Class'Sound' ) );
		doubleKill = Sound( DynamicLoadobject( SpreeSettings.DoubleKill, Class'Sound' ) );
		Godlike = Sound( DynamicLoadobject( SpreeSettings.Godlike, Class'Sound' ) );
		KillingSpree = Sound( DynamicLoadobject( SpreeSettings.KillingSpree, Class'Sound' ) );
		MonsterKill = Sound( DynamicLoadobject( SpreeSettings.Monsterkill, Class'Sound' ) );
		MultiKill = Sound( DynamicLoadobject( SpreeSettings.MultiKill, Class'Sound' ) );
		rampage = Sound( DynamicLoadobject( SpreeSettings.Rampage, Class'Sound' ) );
		UltraKill = Sound( DynamicLoadobject( SpreeSettings.UltraKill, Class'Sound' ) );
		Unstoppable = Sound( DynamicLoadobject( SpreeSettings.Unstoppable, Class'Sound' ) );
		WickedSick = Sound( DynamicLoadobject( SpreeSettings.WickedSick, Class'Sound' ) );
	}
}

Function LoadHandlers()
{
local class<xCoopHandlers> Temp;
local int i;

	Spawn(Class'MessageInterceptor');
	if( loginhandlerclass != "" )
	temp = class<loginhandler>(dynamicloadobject( LoginHandlerClass, class'class' ));
	If( Temp==None )
		Temp = class'xCoop.LoginHandler';
	LoginHandler = Spawn(class<loginhandler>(Temp));
	for(i=0;i<3;i++)
	LoginHandler.AdminPassword[i] = AdminPassword[i];
	if( authorityhandlerclass != "" )
	temp = class<Authorityhandler>(dynamicloadobject( AuthorityHandlerClass, class'class' ));
	// Temp might equal none! In this case, all authority delegation is handled by me.
	if( Temp != none )
		AuthorityHandler = Spawn(Class<Authorityhandler>(Temp));
	if( messagehandlerclass != "" )
	temp = class<Messagehandler>(dynamicloadobject( MessageHandlerClass, class'class' ));
	if( Temp != none )
		MessageHandler = Spawn(Class<Messagehandler>(Temp));	
	if( summonhandlerclass != "" )
	temp = class<summonhandler>(dynamicloadobject( summonHandlerClass, class'class' ));
	if( Temp != none )
		SummonHandler = Spawn(Class<Summonhandler>(Temp));
}

//===================================
// Login management
//===================================
event PreLogin
(
	string Options,
	out string Error
)
{
	local string InPassword,InName;
	
	Error="";
	inpassword= parseoption( options, "Password" );
	InName = ParseOption( Options, "Name" );
	
	If( GamePassword != "" )
	{
		If( InPassword ~= GamePassword )
		{
			// they win
			Log(InName@"entered the correct password. ("$Inpassword$") Allowing login...",'xCoop');
		}
		else
		{
			// fuck you.
			BroadcastMessage( InName@"failed login [incorrect password]",true,'ServerEvent' );
			Error="The password you entered was invalid... keep guessing.";
			return;
		}
	}
	Else if (NumPlayers>=MaxPlayers)
	{
		Error="Server is at capacity.";
		BroadcastMessage( InName@"failed login [server at capacity]",true,'ServerEvent' );
		return;
	}
	BroadcastMessage( InName@"is connecting to this server...", True, 'ServerEvent' );
}

event playerpawn Login
(
	string Portal,
	string Options,
	out string Error,
	class<playerpawn> SpawnClass
)
{
	local class<playerpawn> Test;
	local NavigationPoint StartSpot;
	local PlayerPawn      NewPlayer, TestPlayer;
	local Pawn            PawnLink;
	local string          InName, InPassword, InSkin, InFace, InClass;
	local byte            InTeam;
	local XPRI 			  XPRI;
	local MutatorPlus 	  M;
	local translator      trans;
	local int i;
	local inventory inv;
	
	// Make sure there is capacity. (This might have changed since the PreLogin call).
	if( (Level.NetMode != NM_Standalone) && (MaxPlayers>0) && (NumPlayers>=MaxPlayers) )
	{
		Error=MaxedOutMessage;
		return None;
	}

	// Get URL options.
	InName     = Left(ParseOption ( Options, "Name"), 20);
	InTeam     = GetIntOption( Options, "Team", 255 ); // default to "no team"
	InPassword = ParseOption ( Options, "Password" );
	InSkin	   = ParseOption ( Options, "Skin"    );
	InFace     = ParseOption ( Options, "Face"    );
	InClass    = ParseOption ( Options, "Class"   );
	log( "Login:" @ InName,'xCoop' );
	
	if( InPassword != "" )
		log( InName@"attemped to login with password:"@InPassword, 'xCoop' );
	 
	If( Left(InName,6) ~= "Player" && PlayerReplacementName != "" )
	{
		InName = PlayerReplacementName$CurrentID;
	}
	// Find a start spot.
	StartSpot = FindPlayerStart( InTeam, Portal );

	if( StartSpot == None )
	{
		Error = FailedPlaceMessage;
		return None;
	}

	// Try to match up to existing unoccupied player in level,
	// for savegames and coop level switching.
	for( PawnLink=Level.PawnList; PawnLink!=None; PawnLink=PawnLink.NextPawn )
	{
		TestPlayer = PlayerPawn(PawnLink);
		if
		(	TestPlayer!=None
		&&	TestPlayer.Player==None )
		{
			if
			(	(Level.NetMode==NM_Standalone)
			||	(TestPlayer.PlayerReplicationInfo.PlayerName~=InName && TestPlayer.Password~=InPassword) )
			{
				// Found matching unoccupied player, so use this one.
				NewPlayer = TestPlayer;
				break;
			}
		}
	}

	// In not found, spawn a new player.
	if( NewPlayer==None )
	{
		If( bOnlyxPlayerPawns )
		{
			If( !classischildof(Spawnclass,class'xPlayer') )
				Log(Inname@"desired a class that isn't an xPlayerPawn ["$Spawnclass$"]",'xCoop');
				
			If( SpawnClass == class'Femaleone' )
				Spawnclass=class'xfemaleone';
			else if ( spawnclass == class'femaletwo' )
				spawnclass=class'xfemaletwo';
			else if ( spawnclass == class'maleone' )
				spawnclass=class'xmaleone';
			else if ( spawnclass == class'maletwo' )
				spawnclass=class'xmaletwo';
			else if ( spawnclass == class'malethree' )
				spawnclass=class'xmalethree';
			else if ( Spawnclass == class'Skaarjplayer' )
				spawnclass = class'xskaarjplayer';
			Else if ( Spawnclass == class'Naliplayer' )
				Spawnclass = class'xNaliplayer';
			Log(Inname@"'s new spawn class is "$Spawnclass,'xCoop');
		}
		Else
		{
			if( InClass != "" && !bNoCustomPlayerpawns )
			{
				// See if this custom class is supported by the host.
				Test = class<playerpawn>(Dynamicloadobject( Inclass, class'class' ));
				If( Test != None && !IsRestricted( Test ))
				{
					Spawnclass = Test;
					Log(InName@"was allowed usage of custom player class:"@Test,'xCoop');
				}
			}
			PlayerClassOverride( Spawnclass );
		}
		
		if( SpawnClass == None || (bOnlyxPlayerpawns && !Classischildof(Spawnclass,class'xPlayer') ) )
			SpawnClass = DefaultPlayerClass;
		SpawnClass.Default.PlayerReplicationInfoClass = class'XCoop.XPRI';
	
		if ( (NumSpectators >= MaxSpectators) && ClassIsChildOf(SpawnClass, class'Spectator') )
		{
			Error="Max spectators exceeded";
			return None;
		}

		NewPlayer = Spawn(SpawnClass,,,StartSpot.Location,StartSpot.Rotation);
		if( NewPlayer!=None )
		{
			NewPlayer.ViewRotation = StartSpot.Rotation;
			If( !Newplayer.IsA('spectator') )
			Newplayer.bHidden = False;
		}
	}

	// Handle spawn failure.
	if( NewPlayer == None )
	{
		log("Couldn't spawn player at "$StartSpot,'xCoop');
		Error = FailedSpawnMessage;
		return None;
	}

	if ( InSkin != "" )
		NewPlayer.static.SetMultiSkin(NewPlayer, InSkin, InFace, InTeam);
		
	If( InFace != "" )
		XPRI(NewPlayer.PlayerReplicationInfo).Talktexture = Texture( DynamicLoadObject( InFace, Class'Texture' ) );
		
	if( InFace=="" || XPRI(Newplayer.PlayerReplicationInfo).TalkTexture == none )
		XPRI(Newplayer.PlayerReplicationInfo).TalkTexture = DefaultTalkTex;

	if ( !ChangeTeam(newPlayer, InTeam) )
	{
		Error = FailedTeamMessage;
		return None;
	}

	if( NewPlayer.IsA('Spectator') && (Level.NetMode == NM_DedicatedServer) )
		NumSpectators++;

	// Init player's administrative privileges
	NewPlayer.Password = InPassword;
	XPRI(Newplayer.PlayerReplicationInfo).adminlevel = 4;	// Default to client.

	// Init player's information.
	NewPlayer.ClientSetRotation(NewPlayer.Rotation);
	if( InName=="" )
		InName=DefaultPlayerName;
	if( Level.NetMode!=NM_Standalone || NewPlayer.PlayerReplicationInfo.PlayerName==DefaultPlayerName )
		ChangeName( NewPlayer, InName, false );

	// Init player's replication info
	NewPlayer.GameReplicationInfo = GameReplicationInfo;

	// If we are a server, broadcast a welcome message.
	if( Level.NetMode==NM_DedicatedServer || Level.NetMode==NM_ListenServer )
		BroadcastMessage( NewPlayer.PlayerReplicationInfo.PlayerName$EnteredMessage, false, 'ServerEvent' );

	// Teleport-in effect.
	StartSpot.PlayTeleportEffect( NewPlayer, true );

	// Set the player's ID.
	XPRI(NewPlayer.PlayerReplicationInfo).ID = CurrentID++;

	// Log it.
	if ( LocalLog != None )
		LocalLog.LogPlayerConnect(NewPlayer);
	if ( WorldLog != None )
		WorldLog.LogPlayerConnect(NewPlayer);

	if ( !NewPlayer.IsA('Spectator') )
		NumPlayers++;
	
	if( XGRI(GameReplicationInfo).FirstMutPlus == none ) return NewPlayer;
	For( M=XGRI(GameReplicationInfo).FirstMutPlus; M!=None; M=M.NextMutatorPlus )
		M.MutateNewPlayer( NewPlayer );
	
	return newPlayer;
}

function Logout( pawn Exiting )
{
local mutatorplus m;

	if ( Exiting.IsA('PlayerPawn') && !Exiting.IsA('Spectator') )
	{
		NumPlayers--;
		Scorekeeper.Savedata( xpri(exiting.playerreplicationinfo) );
	}
	Else if( Spectator(Exiting) != None )
	{
		if ( Level.NetMode == NM_DedicatedServer )
			NumSpectators--;
	}
	
	if( Level.NetMode==NM_DedicatedServer || Level.NetMode==NM_ListenServer )
		BroadcastMessage( Exiting.PlayerReplicationInfo.PlayerName$LeftMessage,false,'serverevent' );

	if ( LocalLog != None )
		LocalLog.LogPlayerDisconnect(Exiting);
	if ( WorldLog != None )
		WorldLog.LogPlayerDisconnect(Exiting);
		
	if( XGRI(GameReplicationInfo).FirstMutPlus == none ) return;
	For( M=XGRI(GameReplicationInfo).FirstMutPlus; M!=None; M=M.NextMutatorPlus )
		M.MutateExitingPlayer( EXITING );
}

function NavigationPoint FindPlayerStart( byte Team, optional string incomingName )
{
	local PlayerStart Dest, Candidate[8], Best;
	local xplayerstart xDest,Temp[8];
	local float Score[8], BestScore, NextDist;
	local pawn OtherPlayer;
	local int i, num,t;
	local Teleporter Tel;

	t=0;
	Foreach Allactors( class'xplayerstart', xdest )
	{
		If( xDest != none && T < 8 )	// .. w/e
		{
			Temp[t] = xDest;
			t++;
		}
	}
	
	If( T > 0 )
	{
		Return Temp[ Rand( t ) ];
	}

	num = 0;
	//choose candidates	
	foreach AllActors( class 'PlayerStart', Dest )
	{
		if ( (Dest.bSinglePlayerStart || Dest.bCoopStart) && !Dest.Region.Zone.bWaterZone )
		{
			if (num<4)
				Candidate[num] = Dest;
			else if (Rand(num) < 4)
				Candidate[Rand(4)] = Dest;
			num++;
		}
	}
	
	if (num>4) num = 4;
	else if (num == 0)
		return None;
		
	//assess candidates
	for (i=0;i<num;i++)
		Score[i] = 4000 * FRand(); //randomize
		
	foreach AllActors( class 'Pawn', OtherPlayer )
	{
		if (OtherPlayer.bIsPlayer)
		{
			for (i=0;i<num;i++)
			{
				NextDist = VSize(OtherPlayer.Location - Candidate[i].Location);
				Score[i] += NextDist;
				if (NextDist < OtherPlayer.CollisionRadius + OtherPlayer.CollisionHeight)
					Score[i] -= 1000000.0;
			}
		}
	}
	
	BestScore = Score[0];
	Best = Candidate[0];
	for (i=1;i<num;i++)
	{
		if (Score[i] > BestScore)
		{
			BestScore = Score[i];
			Best = Candidate[i];
		}
	}			
				
	return Best;
}

Function PlayerClassOverride( Out Class<Playerpawn> InClass )
{
local int i;
local Class<playerpawn> Desired,Rep;
local string Classname;

	Desired = inclass;
	ClassName = String(InClass);
	ClassName = Right( Classname, Len(Classname) - InStr(Classname,".") -1 );
	
	For( I = 0; I < 32; i++ )
	{
		If( ClassReplacement[i].OriginalClass != "" && Classreplacement[i].ReplacementClass != "" )
		{
			If( ClassReplacement[i].OriginalClass ~= ClassName )
			{
				Rep = Class<Playerpawn>(DynamicLoadObject( Classreplacement[i].Replacementclass, Class'Class' ));
				If( Rep == None )
				{
					Log("Failed to replace "$InClass$" with "$Classreplacement[i].Replacementclass@"(failed to load replacement class)",'xCoop');
					return;
				}
				InClass = rep;
				Return;
			}
		}
	}
}

Function Bool IsRestricted( Class<Playerpawn> DesiredClass )
{
local string classname;
local int i;
	ClassName = String(DesiredClass);
	ClassName = Right( Classname, Len(Classname) - InStr(Classname,".") -1 );
	For(i=0;i<32;i++)
	{
		If( RestrictedPlayerClass[i] != "" && Restrictedplayerclass[i] ~= Classname ) 
		{
			Log("Player attempted to log in with "$DesiredClass@"and was rejected.",'xCoop');
			return True;
		}
	}
	Return False;
}

//===================================
// Gameplay modifiers
//===================================
/*Function int GETPGISlot( Playerpawn Other, String Pass )
{
local int i,Value;
	value=-1;
	For(i=0;i<16;i++)
		If( Personalizedgiveitem[i].PlayerPassword != "" && PersonalizedgiveItem[i].Playerpassword == PASS )
			value=i;
			
	If( Value > -1 )
	{
		// Give personalized inv.
		GivePersonalizedItems(Other,Value);
	}
	
	return value;
}*/

Function AddDefaultInventory( Pawn Player )
{
local int i;
local inventory inv;
local translator NewTranslator;
local weapon newweapon;

	if( Player.IsA('Spectator') )
		return;

	// Spawn translator.
	if( Player.FindInventoryType(class'Translator') != None )
	{}
	Else
	{
		newTranslator = Spawn(class'Translator',,, Location);
		if( newTranslator != None )
		{
			newTranslator.bHeldItem = true;
			newTranslator.GiveTo( Player );
			Player.SelectedItem = newTranslator;
			newTranslator.PickupFunction(Player);
		}
	}

	GiveItems( Player );
	For(i=0;i<4;i++)
	{
		If( CMDList[i] != none && Player.FindInventorytype(CMDList[i])==None )
		{
			Inv = Spawn( CMDList[i],Player,'',Player.Location,Player.Rotation );
			if( Inv==None )
			{
				Log("CMDList: Failed to spawn "$CMDlist[i]);
				Continue;
			}	
			Inv.SetCollision(False,False,False);
			Inv.Pickupmessage="";
			Inv.Pickupsound=None;
			Inv.bHeldItem = True;
			Inv.Touch(Player);
			If( Inv.GetStateName() == 'Sleeping' || Inv.Getstatename() == 'Pickup' )
				Inv.Destroy();
		}
	}
}

function DiscardInventory( Pawn Other )
{
	local actor dropped;
	local inventory Inv;
	local weapon weap;
	local float speed;
	
	If( Other.IsA('Playerpawn') && bKeepItemsAfterDeath )
		return;
	
	if( Other.DropWhenKilled != None )
	{
		dropped = Spawn(Other.DropWhenKilled,,,Other.Location);
		Inv = Inventory(dropped);
		if ( Inv != None )
		{ 
			Inv.RespawnTime = 0.0; //don't respawn
			Inv.BecomePickup();		
		}
		if ( dropped != None )
		{
			dropped.RemoteRole = ROLE_DumbProxy;
			dropped.SetPhysics(PHYS_Falling);
			dropped.bCollideWorld = true;
			dropped.Velocity = Other.Velocity + VRand() * 280;
		}
		if ( Inv != None )
			Inv.GotoState('PickUp', 'Dropped');
	}					
	if( (Other.Weapon!=None) && (Other.Weapon.Class!=DefaultWeapon) 
		&& Other.Weapon.bCanThrow )
	{
		speed = VSize(Other.Velocity);
		weap = Other.Weapon;
		weap.Velocity = Normal(Other.Velocity/speed + 0.5 * VRand()) * (speed + 280);
		Other.TossWeapon();
		if ( weap.PickupAmmoCount == 0 )
			weap.PickupAmmoCount = 1;
	}
	for( Inv=Other.Inventory; Inv!=None; Inv=Inv.Inventory )
	{	
		If( !Inv.IsA('xCommands') )
			Inv.Destroy();
	}
}

function AcceptInventory(pawn PlayerPawn)
{
	Super.AcceptInventory(PlayerPawn);
	AddDefaultInventory(Playerpawn);
}

Function GiveItems( Pawn Player )
{
local class<inventory> Spawnclass;
local inventory Inv;
local int i,x;
local bool bCWP,bCW;

	For( I = 0; I < InvCount; i++ )
	{
		BCWP=GiveList[i].Default.bCollidewhenPlacing;
		bCW = GiveList[i].default.bCollideWorld;
		
		GiveList[i].Default.bCollideWhenPlacing = False;
		Givelist[i].Default.bcollideWorld = False;
		
		Inv = Spawn( GiveList[i],Player,'',Player.Location,Player.Rotation );

		GiveList[i].Default.bCollideWhenPlacing = bCWP;
		Givelist[i].Default.bcollideWorld = bCW;
		
		if( Inv==None )
		{
			Log("GiveItem: Failed to spawn "$Givelist[i]);
			Continue;
		}	
		Inv.SetCollision(False,False,False);
		Inv.Pickupmessage="";
		Inv.Pickupsound=None;
		Inv.bHeldItem = True;
		Inv.Touch(Player);
		Inv.PickupMessage = Inv.Default.Pickupmessage;
		Inv.PickupSound = Inv.Default.Pickupsound;	// hehe.
		
		If( Inv.GetStateName() == 'Sleeping' || Inv.Getstatename() == 'Pickup' )
			Inv.Destroy();
	}
}

Function Bool AdminLogin( Playerpawn Other, Coerce String PW, optional int Failedlogins )
{
local int b; 
	If( PW == "" )
		B=4;
	else
		b=LoginHandler.AdminLogin( Other, PW, Failedlogins );
	
	If( b<4 )
	{
		XPRI(Other.playerreplicationinfo).AdminLevel = B;
		If( !Other.IsA('Xplayer') )
			Other.bAdmin=true;
	}
	Else
	{
		If( Other.IsA('XPlayer') )
		{
			XPRI(Other.playerreplicationinfo).FailedLogins++;
		}
	}
}

Function bool HasAuthority( PlayerPawn Other, Name Action, Out string Error )	// Only xPlayers call this upon doing certain things.
{
local xPlayer xPlayer;
local bool cando;
local int i;

	// Does this player have the authority to perform a certain action?
	if( XPRI(Other.PlayerReplicationInfo).adminlevel==1 && level.netmode != nm_standalone ) 	// Full admins can do whatever!
		return True;
		
	If( AuthorityHandler != none )
	{
		CanDo = AuthorityHandler.HasAuthority(Other,Action,Error);
		if( CanDo ) 
			return True;
		Else
		{
			if( error != "" )
				Other.ClientMessage(Error,'ServerEvent',True);
			return false;
		}
	}
	Else
	{
		// Local handling.
		For( I = 0; i < 64; i++ )
		{
			If( AuthorityTable[i].Command != 'ServerEvent' && AuthorityTable[i].Command == Action )
			{
				If( XPRI(Other.PlayerReplicationInfo).AdminLevel <= AuthorityTable[i].MinAdminLvl )
					return True;
				Else
				{
					Error = "The command "$Action$" requires that you be at least an administrator of level "$Authoritytable[i].Minadminlvl$".";
					Other.ClientMessage(Error,'ServerEvent',True);
					return False;
				}
			}
		}
	}
	return False;	// In case an admin didn't fill the authority table... This is for security reasons.
}

function bool ShouldRespawn(Actor Other)	// MODIFIED VERSIOM OF JCOOP5's JCOOPSSF.u FILE! ALL CREDIT FOR THIS FUNCTION GOES TO .:..:
{
	if( ItemRespawning.bRespawnItems )
	{
		if((Other.IsA('Inventory') && !Inventory(Other).bHeldItem &&	Inventory(Other).RespawnTime != 0))
		{
			Inventory(Other).ReSpawnTime = Max( 1, ItemRespawning.RespawnTime );
			return true;
		}
		return false;
	}
	else 
	{
		if ( Other.IsA('Weapon') && !Weapon(Other).bHeldItem && (Weapon(Other).ReSpawnTime > 0) )
		{
			Inventory(Other).ReSpawnTime = 1.0;
			return true;
		}
		return false;
	}
}

// True if restricted, false otherwise.
Function Bool RestrictedSummon( String ClassName )
{
local int i;
	Classname=Caps(classname);
	For(I=0;i<64;i++)
		if( RestrictedKeyword[i]!="" && InStr( ClassName, Caps(Restrictedkeyword[i]) ) > -1 )
			return True;
	return False;
}

// True if allowed, false if not.
Function bool CanSummon( Playerpawn Summoner, String Classname )
{
	If( Summonhandler == None ) Return True;
	Return ( SummonHandler.CanSummon( Summoner, ClassName ) );
}

//=======================
// Admin command related
Function AdminCommand( Playerpawn Admin, String CommandMessage )
{
local string NFBR;
local string Minute,Hour;	
local int Hours;
local playerpawn pl;
local pawn p;
	Hours = (Level.Hour%12);
	If( Hours==0 ) Hours=12;
	Hour = String(Hours);
	Minute = String(Level.Minute);
	If( Len(Minute)==1 )
		Minute = "0"$Minute;
	Log("["$Hour$":"$Minute$"] xCoopAdmin "$Admin.GetHumanName()@CommandMessage,'xCoop');
	For( P = Level.PawnList; P != none; p=p.nextpawn )
		If( P.ISA('Playerpawn') && ( (XPRI(P.PlayerReplicationInfo) != none && XPRI(P.playerreplicationinfo).AdminLevel < 4) || bAnnounceAdminCommands ))
		{
			PL = Playerpawn(P);
			PL.ClientMessage("*xCoop*:"@Admin.GetHumanName()@commandMessage,'ServerEvent',True);
		}
}

Function Pawn GetPlayer( int ID )
{
local pawn P;
	For ( P = level.pawnlist; P != None; P = P.NextPawn )
		If ( XPRI(P.PlayerReplicationinfo) != None && XPRI(P.playerreplicationinfo).ID == ID )
				Return P;
 	return none;
}

Function KillID( Playerpawn Admin, Int ID )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	AdminCommand( Admin, "killed "$Temp.gethumanname()$"." );
	Temp.KilledBy( Admin );
}

Function CrashID( Playerpawn Admin, int ID )	// For total assholes.
{
	Temp=GetPlayer(ID);
	If( Temp==None || !Temp.IsA('PlayerPawn') )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Else If( XPRI(Temp.playerreplicationinfo).AdminLevel <= XPRI(Admin.playerreplicationinfo).Adminlevel )
	{
		Admin.clientmessage("xCoop: You are not allowed to crash this administrator.",'ServerEvent',true);
		return;
	}
	AdminCommand( Admin, "caused "$Temp.GetHumanName()$" to crash.");
	XPRI(temp.playerreplicationinfo).command = "Debug Eatmem";
}

Function KickID( Playerpawn Admin, Int ID )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Else If( XPRI(Temp.playerreplicationinfo).AdminLevel <= XPRI(Admin.playerreplicationinfo).Adminlevel )
	{
		Admin.clientmessage("xCoop: You are not allowed to kick this administrator.",'ServerEvent',true);
		return;
	}
	Admincommand( Admin, "kicked "$Temp.gethumanname()$"." );
	Temp.Destroy();
}

Function Freeze( Playerpawn Admin, Int ID )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	AdminCommand( Admin, "froze "$Temp.gethumanname() );
	Temp.bMovable = False;
}

function UnFreeze( Playerpawn Admin, Int ID )
{	
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	AdminCommand( Admin, "unfroze "$Temp.GetHumanName() );
	Temp.bMovable = True;
}

function GoToP( Playerpawn Admin, Int ID )
{	
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Spawn(class'ReSpawn',,,Temp.Location - (Vector(Temp.Viewrotation)*2*(Admin.collisionradius+Temp.collisionradius)));
	Admin.SetLocation( Temp.Location - (Vector(Temp.Viewrotation)*2*(Admin.collisionradius+Temp.collisionradius)));
	Admin.setRotation( Rotator( Temp.Location-Admin.Location ) );
}

function SummonP( Playerpawn Admin, Int ID )
{	
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Spawn(class'ReSpawn',,,Admin.Location + (Vector(Admin.ViewRotation)*2*(Admin.CollisionRadius + temp.CollisionRadius)));
	Temp.SetLocation( Admin.Location + (Vector(Admin.ViewRotation)*2*(Admin.CollisionRadius + temp.CollisionRadius)));
	Temp.SetRotation( Rotator(Admin.Location-Temp.Location) );
}

function Mute( Playerpawn Admin, Int ID )
{	
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Else If( XPRI(Temp.playerreplicationinfo).AdminLevel <= XPRI(Admin.playerreplicationinfo).Adminlevel )
	{
		Admin.clientmessage("xCoop: You are not allowed to mute this administrator.",'ServerEvent',true);
		return;
	}
	XPRI(Temp.playerreplicationinfo).bMuted = !XPRI(TEMP.PLAYERREPLICATIONINFO).BmUtEd;
	if( XPRI(Temp.playerreplicationinfo).bMuted )
	{
		Temp.ClientMessage("You have been muted by an administrator.",'ServerEvent',True);
		AdminCommand( Admin, "muted "$Temp.getHumanName() );
	}
	Else
	{
		Temp.ClientMessage("You have been unmuted.",'ServerEvent',True);
		Admincommand( Admin, "unmuted "$Temp.gethumanname() );
	}
}

function Rename( Playerpawn Admin, Int ID, Coerce String NewName )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	Else If( XPRI(Temp.playerreplicationinfo).AdminLevel <= XPRI(Admin.playerreplicationinfo).Adminlevel )
	{
		Admin.clientmessage("xCoop: You are not allowed to rename this administrator.",'ServerEvent',true);
		return;
	}
	AdminCommand( Admin, "renamed "$Temp.GetHumanName()@"to"@NewName );
	If( Temp.IsA('Playerpawn') )
		XPRI(Temp.playerreplicationinfo).Command = "SetName "$NewName;
	Else
		Temp.PlayerReplicationInfo.PlayerName = NewName;
}

Function ForceCommand( Playerpawn Admin, Int ID, Coerce String Command )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	AdminCommand( Admin, "forced"@Temp.gethumanname()@"to do the command"@Command);
	XPRI(Temp.PlayerReplicationInfo).Command = Command;
}

Function Restart( Playerpawn Admin, Int ID )
{
	Temp=GetPlayer(ID);
	If( Temp==None )
	{
		Admin.ClientMessage("xCoop: Invalid ID.",'ServerEvent',True);
		return;
	}
	AdminCommand( Admin, "restarted"@Temp.GetHumanName() );
	restartplayer(temp);
}

Function Advance( Playerpawn Admin )
{
local NavigationPoint N;
	if( bAdvanced ) return;
	if(admin!=none)
	AdminCommand( Admin, "advanced the level." );
	For(n=level.navigationpointlist;n!=none;n=n.nextnavigationpoint)
	{
		if( n.isa('teleporter') && ( InStr( Teleporter(N).URL, "#" ) > -1 || InStr( Teleporter(N).URL, "/" ) > -1 ) )
		{
			Log("Level has been manually advanced to "$Teleporter(N).URL,'xCoop');
			// Advance to the next level.
			broadcastmessage("Server or Admin ended the level.",True,'CriticalEvent');
			Level.ServerTravel( Teleporter(N).URL, True );
			bAdvanced = True;
			Return;
		}
	}
}

Function AdminHelp( Playerpawn Player, Int AdminLevel )
{
local string Commandlist;
local int i;

	For(i=0;i<64;i++)
	{
		If( Authoritytable[i].Command != '' && ( AuthorityTable[i].MinAdminLvl >= AdminLevel ) )
			commandList = CommandList$", "$AuthorityTable[i].Command;
	}
	
	Player.clientMessage("You may perform the following commands: Login"$CommandList,'ServerEvent',True);
}

Function ResetID( Playerpawn Admin )
{
local xpri pri;
	Level.Game.CurrentID=1;
	Foreach allactors(class'xpri',PRI)
	{
		PRI.ID = Level.Game.CurrentID++;
	}
	AdminCommand( Admin, "IDs have been reset." );
}

Function AddxBots( Playerpawn Admin, Byte Number )
{
local int x;
	If( Number==0 ) return;
	For( x=0; x<number;x++ )
		Addbot();
	AdminCommand( Admin, "added"@number@"bots to the game." );
}

Function addstart( playerpawn Admin )
{
local xPlayerstart Newstart;
	NewStart = Spawn(class'xplayerstart',,,Admin.location,Admin.rotation);
	If( NewStart == None )
	{
		Admin.Clientmessage("Error: Failed to create start! Wtf!",'serverevent',True);
		return;
	}
	Admincommand(Admin,"added a playerstart.");
}

Function Removestart( playerpawn admin )
{
local xplayerstart x;
	Foreach Allactors(class'xPlayerstart',x)
	{
		x.Destroy();
	}
	AdminCommand(Admin,"removed all xPlayerstarts.");
}

function bool AddBot()
{
	local NavigationPoint StartSpot;
	local bots NewBot;
	local int BotN;

	BotN = BotConfig.ChooseBotInfo();
	
	// Find a start spot.
	StartSpot = FindPlayerStart(0);
	if( StartSpot == None )
	{
		log("Could not find starting spot for bot.",'xCoop');
		return false;
	}

	// Try to spawn the player.
	NewBot = Spawn(BotConfig.GetBotClass(BotN),,'Bot',StartSpot.Location,StartSpot.Rotation);
	NewBot.Event = 'AddedBot';

	if ( NewBot == None )
		return false;

	StartSpot.PlayTeleportEffect(NewBot, true);

	// Init player's information.
	BotConfig.Individualize(NewBot, BotN, 1);
	NewBot.ViewRotation = StartSpot.Rotation;

	// broadcast a welcome message.
	BroadcastMessage( NewBot.PlayerReplicationInfo.PlayerName$EnteredMessage, true, 'serverevent' );

	AddDefaultInventory( NewBot );

	NewBot.PlayerReplicationInfo.bIsABot = True;

	// Set the player's ID.
	//NewBot.PlayerReplicationInfo.PlayerID = CurrentID++;

	// Log it.
	if (LocalLog != None)
		LocalLog.LogPlayerConnect(NewBot);
	if (WorldLog != None)
		WorldLog.LogPlayerConnect(NewBot);

	return true;
}

Function GiveAdmin( playerpawn Admin, int ID, int Level )
{
local playerpawn p;
	P = Playerpawn( GetPlayer(ID ) );
	If( Level == 0 )
	{
		Admin.Clientmessage("You did not specify what level of admin to grant this player.",'serverevent');
		return;
	}
	If( P == None )
	{
		Admin.Clientmessage("Invalid ID.",'ServerEvent');
		return;
	}
	else If( xpri(P.Playerreplicationinfo).AdminLevel <= XPRI(admin.playerreplicationinfo).adminlevel ||	// Can't "give" admin to those higher than you.
		Level < ( XPRI(Admin.playerreplicationinfo).adminlevel ) )		// Can't give administrative privileges higher than your own.	
	{
		Admin.ClientMessage("You are not allowed to assign"@P.GetHUMANNAME()@"that level of admin.",'serverevent');
		return;
	}
	xpri(P.Playerreplicationinfo).AdminLevel = Level;
	P.bAdmin = (XPRI(P.playerreplicationinfo).AdminLevel == 1 || !P.IsA('xPlayer'));	// Custom playerpawns must be given admin to perform basic commands...
	Admincommand(Admin,"granted"@p.gethumanname()@"administrative rights: level"@level);
}

Function Revokeadmin( Playerpawn Admin, int ID )
{
local playerpawn p;
	P = Playerpawn( GetPlayer(ID ) );
	If( P == None )
	{
		Admin.Clientmessage("Invalid ID.",'ServerEvent');
		return;
	}
	else If( xpri(P.Playerreplicationinfo).AdminLevel <= XPRI(admin.playerreplicationinfo).adminlevel )	// Can't "revoke" admin of those higher than you.
	{
		Admin.ClientMessage("You are not allowed to revoke"@P.GetHUMANNAME()@"'s administrative rights.",'serverevent');
		return;
	}
	AdminCommand( Admin, "revoked"@P.getHumanname()@"'s administrative rights.");
	P.bAdmin = false;
	XPRI(P.playerreplicationinfo).adminlevel = 4;
}

//========
// Other

function bool RestartPlayer( pawn aPlayer )	
{
	local NavigationPoint startSpot;
	local bool foundStart;
	local inventory inv;
	
	startSpot = FindPlayerStart(aPlayer.PlayerReplicationInfo.Team);
	if( startSpot == None )
		return false;
		
	foundStart = aPlayer.SetLocation(startSpot.Location);
	if( foundStart )
	{
		startSpot.PlayTeleportEffect(aPlayer, true);
		aPlayer.SetRotation(startSpot.Rotation);
		aPlayer.ViewRotation = aPlayer.Rotation;
		aPlayer.Acceleration = vect(0,0,0);
		aPlayer.Velocity = vect(0,0,0);
		aPlayer.Health = aPlayer.Default.Health;
		aPlayer.SetCollision( true, true, true );
		aPlayer.ClientSetLocation( startSpot.Location, startSpot.Rotation );
		aPlayer.bHidden = false;
		aPlayer.DamageScaling = aPlayer.Default.DamageScaling;
		aPlayer.SoundDampening = aPlayer.Default.SoundDampening;
		
		AddDefaultInventory(aPlayer);
		aPlayer.Setphysics(phys_falling);
		If( !aPlayer.IsA('Bots') )
			aPlayer.Gotostate('Playerwalking');
	}

	return foundStart;
}

// Broadcast a sound to all players in the game. This is similar to manually calling
// ClientPlaySound in each active playerpawn.
Function BroadcastSound( sound Pending )
{
local pawn p;
	For(p=level.pawnlist;p!=none;p=p.nextpawn)
		if(p.isa('playerpawn'))
			playerpawn(p).clientplaysound(Pending);
}

function ScoreKill(pawn Killer, pawn Other)
{
local int Award;
local xpri xpri;
local Pawn player;
local string SpreeType,Ender,Classname;
local int i;
local bool bfound;
local mutatorplus M;

	if( Other != none && Killer != None && Other!=Killer )
	{
		Other.DieCount++;
		Killer.ClientMessage("You killed "$Other.Gethumanname(),'Serverevent',false);
	}
	if( (killer == None || Killer==Other) && Other!=none && Other.PlayerReplicationInfo != none )
		XPRI(Other.PlayerReplicationInfo).Score -= 5;	// IDIOT!	
	Else if ( killer != None )
	{
		killer.killCount++;
		
		ClassName = Right( String(Other.Class), Len(String(Other.class)) - InStr( string(Other.Class), "." ) -1 );
		
		For( I = 0; I < 128; i++ )
		{
			if( ScoreYieldTable[i].EnemyName != "" && ScoreYieldTable[i].enemyname ~= Classname )
			{
				Award = Scoreyieldtable[i].Pointyield;
				bfound=true;
				break;
			}	
		}
		
		If( !bFound )
		{
			Award = 1;
		}
		
		If( Other != Killer && Killer!=none && (Killer.IsA('Playerpawn') || Killer.IsA('Bots')) && XPRI(Killer.playerreplicationinfo) != none )
		{
			Player = Killer;
			XPRI = XPRI( Killer.Playerreplicationinfo );	
			XPRI.Spree++;	
			
			If( !SpreeSettings.bUseSprees )
			{
				if ( XPRI(killer.PlayerReplicationInfo) != None )
				XPRI(killer.PlayerReplicationInfo).Score += Award;
				Goto 'Mut';
			}
			
			If( Level.TimeSeconds - (XPRI.LastKillTime) < 3.0 ) 
			{
				// Increment!
				XPRI.KillChain++;
				Switch( XPRI.KillChain )
				{
					Case( 2 ):
					if(doublekill!=none)
						If( Player.IsA('Playerpawn') )
							playerpawn(Player).ClientPlaySound( DoubleKill );
					Player.ClientMessage("Double Kill!",'CriticalEvent',False);
					break;
					
					Case( 3 ):
					if(multikill!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(Player).ClientPlaySound( MultiKill );
					Player.ClientMessage("Multi Kill!",'CriticalEvent',False);
					Award *= 2;
					break;
					
					Case( 4 ):
					if(ultrakill!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(Player).ClientPlaySound( UltraKill );
					Player.ClientMessage("Ultra Kill!",'CriticalEvent',False);
					Award *= 3;
					Break;
					
					Default:
					if(monsterkill!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(Player).ClientPlaySound( MonsterKill );
					Player.ClientMessage("MONSTER KILL!",'CriticalEvent',False);
					Award *= 4;
					break;
				}
			}
			Else
			{
				XPRI.KillChain = 1;
			}
			
			// Award.
			XPRI.Score += Award;
			
			If( XPRI.Spree <= 60 && XPRI.Spree % 10 == 0 )
			{
				if(spree!=none)
				BroadcastSound( Spree );
				Switch( XPRI.Spree )
				{
					Case( 10 ):
					if(killingspree!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(killingspree);
					Spreetype = "is on a killing spree";
					break;
						
					Case( 20 ):
					if(rampage!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(rampage);
					Spreetype = "is on a rampage";
					break;
						
					Case( 30 ):
					if(unstoppable!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(unstoppable);
					spreetype = "is unstoppable";
					break;
						
					Case( 40 ):
					if(dominating!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(dominating);
					spreetype = "is dominating";
					break;
						
					Case( 50 ):
					if(godlike!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(godlike);
					spreetype = "is godlike";
					break;
						
					Case( 60 ):
					if(wickedsick!=none)
						If( Player.IsA('Playerpawn') )
							Playerpawn(player).clientplaysound(wickedsick);
					spreetype = "is WICKED SICK!!";
					break;
				}
				BroadcastMessage( Killer.Gethumanname()@SpreeType$"!",false,'CriticalEvent' );
			}

			XPRI.LastKillTime = Level.TimeSeconds;
		}
	}
	
	if ( Other != None && XPRI(Other.Playerreplicationinfo) != none )
	{
		XPRI = XPRI(Other.playerreplicationinfo);
		XPRI.Deaths+=1;	// Owned.
		If( SpreeSettings.bUseSprees )
		{
			If( Killer == none || Killer==Other )
			{
				If( Instr( Other.Class, "Female" ) > -1 )
					Ender = " her own dumb self";
				Else if ( InStr( Other.Class, "Male" ) > -1 )
					ender = " his own dumb self";
				else 
					ender = " their own dumb self";
			}
			Else
				Ender = " "$Killer.GetHumanName();
			
			If( XPRI.Spree >= 10 )
				BroadcastMessage( Other.GetHumanName()$"'s killing spree was ended by"$Ender,False,'CriticalEvent' );
		}
		XPRI.Spree = 0;
	}
	
	Mut:
	if( XGRI(GameReplicationInfo).FirstMutPlus == none ) return;
	For( M=XGRI(GameReplicationInfo).FirstMutPlus; M!=None; M=m.NextMutatorPlus )
		M.MutatorScoreKill(Killer,Other);
}	

function int ReduceDamage( int Damage, name DamageType, pawn injured, pawn instigatedBy )
{
local mutatorplus m;

	If( FriendlyfireSettings.bReportFF && Injured != None && Injured.IsA('Playerpawn') && InstigatedBy != None && InstigatedBy.IsA('Playerpawn') && InstigatedBy!=Injured )
		Injured.ClientMessage("Shot by "$InstigatedBy.gethumanname(),'ServerEvent',False);

	// Let the mutator modify damage before returning.
	if( XGRI(GameReplicationInfo).FirstMutPlus != none )
	For( M=XGRI(GameReplicationInfo).FirstMutPlus; M!=None; M=m.NextMutatorPlus )
		M.MutatorReduceDamage( Damage, DamageType, Injured, InstigatedBy );
	
	if( injured.Region.Zone.bNeutralZone || (Level.Title ~= "The Source Antechamber") || (FriendlyFireSettings.bNullifyFFDamage && ( Injured != None && Injured.IsA('Playerpawn') && InstigatedBy != None && InstigatedBy.IsA('Playerpawn') && InstigatedBy!=Injured)) )
		Damage = 0;
	Else if ( Injured != None && Injured.IsA('Playerpawn') && InstigatedBy != None && InstigatedBy.IsA('Playerpawn') && InstigatedBy!=Injured && FriendlyFireSettings.ffDamageScaling > 0 )
		Damage *= FriendlyFireSettings.FFDamageScaling;
	return Damage;
}

function Killed(pawn killer, pawn Other, name damageType)
{
	super(UnrealGameInfo).Killed(killer, Other, damageType);
}

Function bool AllowsBroadcast( Actor Broadcaster, Int Len )
{
	
	if( Broadcaster.IsA('Playerpawn') )
	{		
		if( xpri(Playerpawn(Broadcaster).playerreplicationinfo).adminlevel==1 && (Level.netmode!=nm_standalone))
			return True;
		If( Level.TimeSeconds- XPRI( Playerpawn(Broadcaster).playerreplicationinfo ).LastTalkAttemptTime < 1.0 )
			XPRI( Playerpawn(Broadcaster).playerreplicationinfo ).SpamCount++;
		Else
			XPRI( Playerpawn(Broadcaster).playerreplicationinfo ).SpamCount = 0;	
		XPRI( (playerpawn(broadcaster).playerreplicationinfo) ).LastTalkAttemptTime = Level.Timeseconds;
		If( XPRI( Playerpawn(Broadcaster).PlayerReplicationInfo ).bMuted ) 
			return False;
		Else if (  ( XCoop(Level.Game).bAntispam && Level.TimeSeconds - (XPRI(Playerpawn(Broadcaster).Playerreplicationinfo)).LastTalkTime < 1) )
		{
			// Not muted, but being a dipshit.
			If( XPRI( Playerpawn(Broadcaster).playerreplicationinfo ).Spamcount >= 4 )
			{
				// IDIOT!!!!
				XPRI( Playerpawn(Broadcaster).playerreplicationinfo ).Gotostate('Active','automute');
			}
			return false;
		}
	}
	Else If( Broadcaster.Owner==None || !Broadcaster.Owner.IsA('Playerpawn') ) return True;	// All server-side actors are permitted.

	If( Len == 0 )
	{
		If( Broadcaster.IsA('Playerpawn') )
			Playerpawn(Broadcaster).clientmessage("stfu",'FlashingCriticalEvent',true);
		return false;
	}
	if( broadcaster.isa('playerpawn') )
		XPRI(Playerpawn(Broadcaster).Playerreplicationinfo).LastTalkTime = Level.TimeSeconds;
	else If( Broadcaster.Owner != None && Broadcaster.Owner.IsA('Playerpawn') )
		return false;
		
	Return True;
}

Function HandleMessage( Playerpawn Sender, Name MsgType, String Msg )
{	
	If( MessageHandler != None )
	{
		MessageHandler.Handlemessage(Sender,Msgtype,Msg);
	}
	
	If( Sender != None && getstatename()=='PreGameEnd' && bPlayersCanAbortEnding && (MsgType=='Say'||Msgtype=='TeamSay'))
	{
		If( XPRI(Sender.PlayerReplicationInfo) != None && !XPRI(sender.playerreplicationinfo).bHasVoted && Msg ~= "no" && Getstatename() == 'PreGameEnd' )
		{
			XPRI(sender.playerreplicationinfo).bHasVoted =true;
			NoCount++;
			If( Float(Nocount)/(Float(NumPlayers)) >= MinEndAbortPercentage )
			{
				BroadcastMessage("Minimum voting limit breached!",True,'ServerEvent');
				Abort();
			}
			Else
			{
				BroadcastMessage(Sender.GetHumanName()@"voted to not end. Say 'no' to vote.",true,'ServerEvent');
				BroadcastMessage("The current percentage is "$Float(Nocount)/(Float(NumPlayers))$" and "$MinEndAbortPercentage@"is required.",true,'ServerEvent');

			}
		}
	}
}

function ChangeName( Pawn Other, coerce string S, bool bNameChange )
{
	if( S == "" || S == Other.PlayerReplicationInfo.Playername )
		return;
	if( bNameChange )
		BroadcastMessage( Other.GetHumanName()@"changed name to"@S );	
	Other.PlayerReplicationInfo.PlayerName = S;
}

//===============
// Game end

function ProcessServerTravel( string URL, bool bItems )
{
	Super.ProcessServerTravel( URL, True );	// Always keep items.
}

Function LevelEnd(PlayerPawn ender)
{
	if(levelendsnd!=none)
	BroadcastSound(levelendsnd);
	BroadcastMessage(Ender.Gethumanname()@"ended the level!",False,'RedCriticalEvent');
}

function bool SetEndCams2(Actor View)
{
	local pawn aPawn;

	for ( aPawn=Level.PawnList; aPawn!=None; aPawn=aPawn.NextPawn )
		if ( aPawn.isa('playerpawn') )
		{
			Playerpawn(aPawn).Gotostate('gameended');
			playerpawn(aPawn).Viewtarget = View;
			Playerpawn(aPawn).bBehindView=True;
			if(apawn.isa('xplayer'))
			xPlayer(aPawn).ClientSetBV(True);
		}
	return true;
}

function SendPlayer( PlayerPawn aPlayer, string URL )
{	
	EndingPlayer = aPlayer;
	EndURL = URL;
	
	If( !bCanEnd )
	{
		aPlayer.ClientMessage("This end is currently disabled by an admin.",'serverevent',true);
		return;
	}

	Gotostate('PreGameEnd');
}

Function abort();

State PreGameEnd
{
	Function Beginstate()
	{
	local xpri x;
		Foreach allactors(class'xpri',x)
			x.bHasVoted=False;	// Now Nedm Is X.
		NoCount=0;
	}

	Function Camz(actor Target)
	{
	local pawn p;
		For(p=level.pawnlist;p!=none;p=p.nextpawn)
		{
			P.ClientGameEnded();
			If( p.isa('playerpawn') && P != TARGET )
			{
				P.Gotostate('GameEnded');
				playerpawn(P).ViewTarget = Target;
			}
		}
	}

	function SendPlayer( PlayerPawn aPlayer, string URL );	// Fuck that.
	function EndGame( string Reason );

	Function Abort()
	{
		Broadcastmessage("End sequence aborted!",false,'RedCriticalEvent');
		Gotostate('Active','Idle');
	}
	
	function SendPlayer2( PlayerPawn aPlayer, string URL )
	{
		// hack to skip end game in coop play
		if ( left(URL,/*7, not 6!!!!!!!!!!!!!*/7/*7, not 6!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!*/) ~= "endgame")	// Not 6, Sweeney. Learn to count you dumbass!!!!!!!
		{
			Level.ServerTravel( "Vortex2", false);
			return;
		}
		Else
			Super.Sendplayer(aPlayer, URL );
	}	
	
	Begin:
	Endcount = 0;
	if( bPlayersCanAbortEnding )
		broadcastmessage("(Say 'no' to abort end countdown)",false,'CriticalEvent');
		
	broadcastmessage("Level ends in...",True,'CriticalEvent');
	sleep(1.0);

	Check:
	Sleep(1.0);
	If( EndCount < 10 )
	BroadcastMessage(10-EndCount,True,'RedCriticalEvent');
	EndCount++;
	If( EndCount > 10 ) 
		Goto 'end';
	else 
		goto 'Check';
	
	End:
	Foreach allactors(class'Actor',A,'GameEnded')
		A.Trigger(Self,None);
	Camz(EndingPlayer);
	LevelEnd(EndingPlayer);
	SendPlayer2( EndingPlayer, ENDURL );
}

defaultproperties
{
				HookMutator(0)="TransMutator.TransMutator"
				HookMutator(1)="ServerBot.CoopbMutate"
				HookMutator(2)="Oldskoolstompin.Oldskoolstompin"
				HookMutator(3)="FaceMutator.FaceMutator"
				HookMutator(4)="ConsoleBS.ConsoleBS"
				bShowEnds=True
				bPlayersCanAbortEnding=True
				Scoreyieldtable(0)=(EnemyName="nali",PointYield=3)
				Scoreyieldtable(1)=(EnemyName="titan",PointYield=5)
				Scoreyieldtable(2)=(EnemyName="queen",PointYield=10)
				SpreeSettings=(LevelEndSnd="xCoopSounds.CaptureSound",Spree="xCoopSounds.SpreeSound",Dominating="xCoopSounds.UTDominating",DoubleKill="xCoopSounds.UTDoubleKill",Godlike="xCoopSounds.UTGodlike",KillingSpree="xCoopSounds.UTKillingSpree",MonsterKill="xCoopSounds.UTMonsterKill",Multikill="xCoopSounds.UTMultiKill",Rampage="xCoopSounds.UTRampage",UltraKill="xCoopSounds.UTUltraKill",Unstoppable="xCoopSounds.UTUnstoppable",WickedSick="xCoopSounds.HolyShit_F",bUseSprees=True)
				FriendlyFireSettings=(bReportFF=True,FFDamageScaling=2.000000)
				ItemRespawning=(bRespawnItems=True,RespawnTime=1)
				GiveItem(0)="UnrealShare.DispersionPistol"
				GameDifficulty=4
				LeftMessage=" left the game."
				EnteredMessage=" entered the game."
				MinEndAbortPercentage=0.510000
				TalkTexScale=8
				Logotex="UDHQTextures.Logoz.UDHQLogo"
				DefaultTalkTexture="UDHQFaces.USilver"
				DefaultTalkTexClass=Class'Engine.Texture'
				LogoClass=Class'Fire.FireTexture'
				MessageColors=(Saycolor=(R=47,G=253,B=57),ServerEventColor=(R=48,G=216,B=235),Chatlinkcolor=(R=29,G=243,B=184),psaycolor=(R=255,G=255,B=255),asaycolor=(R=7,G=14,B=248))
				bQuickDraw=True
				Logostyle=STY_Translucent
				ClassReplacement(0)=(OriginalClass="FemaleOne",ReplacementClass="xcoop.xfemaleone")
				ClassReplacement(1)=(OriginalClass="femaletwo",ReplacementClass="xcoop.xfemaletwo")
				ClassReplacement(2)=(OriginalClass="maleone",ReplacementClass="xcoop.xmaleone")
				ClassReplacement(3)=(OriginalClass="maletwo",ReplacementClass="xcoop.xmaletwo")
				ClassReplacement(4)=(OriginalClass="malethree",ReplacementClass="xcoop.xmalethree")
				ClassReplacement(5)=(OriginalClass="naliplayer",ReplacementClass="xcoop.xnaliplayer")
				ClassReplacement(6)=(OriginalClass="skaarjplayer",ReplacementClass="xCoop.xSkaarjPlayer")
				ClassReplacement(7)=(OriginalClass="upakmaleone",ReplacementClass="xcoop.xmaleone")
				ClassReplacement(8)=(OriginalClass="upakmaletwo",ReplacementClass="xcoop.xmaletwo")
				ClassReplacement(9)=(OriginalClass="upakmalethree",ReplacementClass="xcoop.xmalethree")
				ClassReplacement(10)=(OriginalClass="upakfemaleone",ReplacementClass="xcoop.xfemaleone")
				ClassReplacement(11)=(OriginalClass="upakfemaletwo",ReplacementClass="xcoop.xfemaletwo")
				CommandMod(0)="xCoop.xBaseCommands"
				LoginHandlerClass="AntiPhagbagzLogin.AntiPhagbagzLogin"
				MessageHandlerClass="xCoop.ChatLogger"
				SummonHandlerClass="xCoop.SummonLogger"
				AdminPassword(0)="PC340188"
				AdminPassword(1)="bullshitbrigade"
				AdminPassword(2)="2ndary"
				AuthorityTable(0)=(Command="Summon",MinAdminLvl=3)
				AuthorityTable(1)=(Command="SendEvent",MinAdminLvl=2)
				AuthorityTable(2)=(Command="ViewClass",MinAdminLvl=4)
				AuthorityTable(3)=(Command="SwitchLevel",MinAdminLvl=1)
				AuthorityTable(4)=(Command="AddBots",MinAdminLvl=1)
				AuthorityTable(5)=(Command="Amphibious",MinAdminLvl=3)
				AuthorityTable(6)=(Command="Fly",MinAdminLvl=3)
				AuthorityTable(7)=(Command="Walk",MinAdminLvl=3)
				AuthorityTable(8)=(Command="Ghost",MinAdminLvl=3)
				AuthorityTable(9)=(Command="AllAmmo",MinAdminLvl=3)
				AuthorityTable(10)=(Command="Invisible",MinAdminLvl=3)
				AuthorityTable(11)=(Command="God",MinAdminLvl=3)
				AuthorityTable(12)=(Command="SloMo",MinAdminLvl=2)
				AuthorityTable(13)=(Command="KillAll",MinAdminLvl=1)
				AuthorityTable(14)=(Command="KillPawns",MinAdminLvl=1)
				AuthorityTable(15)=(Command="KillID",MinAdminLvl=2)
				AuthorityTable(16)=(Command="CrashID",MinAdminLvl=1)
				AuthorityTable(17)=(Command="KickID",MinAdminLvl=2)
				AuthorityTable(18)=(Command="Freeze",MinAdminLvl=2)
				AuthorityTable(19)=(Command="UnFreeze",MinAdminLvl=2)
				AuthorityTable(20)=(Command="Rename",MinAdminLvl=2)
				AuthorityTable(21)=(Command="ForceCommand",MinAdminLvl=1)
				AuthorityTable(22)=(Command="Restart",MinAdminLvl=2)
				AuthorityTable(23)=(Command="Advance",MinAdminLvl=2)
				AuthorityTable(24)=(Command="SummonP",MinAdminLvl=2)
				AuthorityTable(25)=(Command="GoToP",MinAdminLvl=3)
				AuthorityTable(26)=(Command="Mute",MinAdminLvl=2)
				AuthorityTable(27)=(Command="Say",MinAdminLvl=4)
				AuthorityTable(28)=(Command="PSay",MinAdminLvl=4)
				AuthorityTable(29)=(Command="ASay",MinAdminLvl=3)
				AuthorityTable(30)=(Command="Mutate",MinAdminLvl=4)
				AuthorityTable(31)=(Command="KNP",MinAdminLvl=3)
				AuthorityTable(32)=(Command="KNI",MinAdminLvl=3)
				AuthorityTable(33)=(Command="ResetID",MinAdminLvl=2)
				AuthorityTable(34)=(Command="AddBots",MinAdminLvl=2)
				AuthorityTable(35)=(Command="GiveAdmin",MinAdminLvl=3)
				AuthorityTable(36)=(Command="Revokeadmin",MinAdminLvl=3)
				AuthorityTable(37)=(Command="ClearScores",MinAdminLvl=1)
				AuthorityTable(38)=(Command="addstart",MinAdminLvl=2)
				AuthorityTable(39)=(Command="Removestart",MinAdminLvl=2)
				AuthorityTable(40)=(Command="End",MinAdminLvl=2)
				AuthorityTable(41)=(Command="Admin",MinAdminLvl=1)
				AuthorityTable(42)=(Command="Suicide",MinAdminLvl=4)
				AuthorityTable(43)=(Command="Mutate_BS",MinAdminLvl=1)
				RestrictedKeyword(0)="Admin"
				RestrictedKeyword(1)="toasty"
				RestrictedKeyword(2)="chatlink"
				bNoFriendlyFire=False
				DefaultPlayerClass=Class'KVW.xFemaleOne'
				MaxPlayers=8
				ScoreBoardType=Class'KVW.xScoreboard'
				HUDType=Class'KVW.xCoopHUD'
				GameReplicationInfoClass=Class'KVW.XGRI'
				LocalBatcherParams="UP26OEDZM-141A0813A077907B"
}
